!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!! Copyright (C) 2007 Jason D. Nielsen
!!!
!!! This program is free software! you can redistribute it and/or modify
!!! it under the terms of the GNU General Public License as published by
!!! the Free Software Foundation! either version 2 of the License, or
!!! (at your option) any later version.
!!!
!!! This program is distributed in the hope that it will be useful,
!!! but WITHOUT ANY WARRANTY! without even the implied warranty of
!!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!! GNU General Public License for more details.
!!!
!!! You should have received a copy of the GNU General Public License
!!! along with this program! if not, write to the Free Software
!!! Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

include 'depends.f90'

module fm_NHPP_mod
  use mytype
  implicit none
  private
  public:: fmNHPP
  real(dp), save, allocatable:: Xmat(:,:), Bmat(:,:), time_grid(:), nij(:,:), muij(:,:), &
       & pDVD(:,:), DVD(:,:), Scr(:,:), BBB(:,:), lognij(:), XX(:,:), ipDVD(:,:), Smat(:,:), &
       & mui(:), ni(:), dH(:), Hss(:,:), Hss2(:,:), fDVD(:,:,:)
  real(dp), save, allocatable, target:: coef_t(:)
  real(dp), save:: eff_df, dof
  real(dp), pointer:: beta(:) => null(), psi(:) => null(), spar(:) => null(), tau => null(), gamma => null()
  logical, save, allocatable:: nijNA(:,:)
  integer, save, allocatable:: sqpan(:), pnind(:,:)
  integer, save:: nind, npan, ntp, ntrt, nrep, nXr, nstp, nknt, nBc, ncf, err, ord
  integer, parameter:: maxit=500

contains

subroutine initialize(NData,Xm,Panel,gmm,npn,nsub,ntr,nkn,nst)
  use mytype
  use matrix, only: operator(.kp.), operator(.mm.), operator(.tmm.)
  use fm_util, only: tkspline
  implicit none
  integer, intent(in):: npn, nsub, ntr, nkn, nst
  real(dp), intent(in):: NData(npn,nsub), Panel(npn), gmm
  real(dp), intent(in):: Xm(nsub,ntr)
  integer:: i
  real(dp):: tpanel(npn), expnd(npn+nind), d_tpanel(npn), m_tpanel(npn), tdiff(npn*nsub), &
       & BB(npn,nkn+3), knt(nkn)
  err=0
  expnd=1.0_dp
  nind=nsub
  npan=npn
  ntrt=ntr
  ntp=nint(Panel(npan))
  nXr=npan*nind
  ord=3
  nBc=nkn+ord
  nknt=nkn
  nstp=nst
  ncf=ntrt+nBc-1
  allocate(Bmat(ntp,nBc), time_grid(ntp), coef_t(ncf+nstp+5), nij(npan,nind), &
       & muij(npan,nind), nijNA(npan,nind), DVD(ncf,ncf), pDVD(ncf,ncf), &
       & Scr(ncf,nind), sqpan(npan), BBB(ncf,ncf), lognij(nXr), XX(nXr,ncf), &
       & pnind(2,npan), ipDVD(ncf,ncf), Smat(nknt,nstp+2), ni(nind), mui(nind), &
       & dH(nind), Hss(nstp+2,nstp+2), Hss2(nstp+2,nstp+2), fDVD(nind,ncf,ncf))
  if (ntrt > 1) then
     allocate(Xmat(nind,ntrt-1))
  end if
  nij=NData
  tpanel=Panel/Panel(npn)
  nijNA=.FALSE.
  where(nij< 0)
     nijNA=.TRUE.
     nij=0.0_dp
  end where  
  ni=sum(nij,1)
  dof=gmm
  sqpan=(/(i,i=1,npan)/)
  if (ntrt > 1) then
     Xmat=Xm(:,2:ntrt)
  end if
  d_tpanel=(tpanel-(/0.0_dp,tpanel(1:(npan-1))/))
  m_tpanel=(tpanel+(/0.0_dp,tpanel(1:(npan-1))/))/2.0_dp
  time_grid=(/(i,i=1,ntp)/)
  time_grid=time_grid/Panel(npn)
  call tkspline(time_grid,ntp,ord-1,nknt,Bmat)
  pnind(1,:)=(/1,(nint(Panel(1:(npan-1)))+1)/)
  pnind(2,:)=nint(Panel)

  knt=(/(real(i,dp),i=1,nknt)/)
  knt=knt/real(nknt+1,dp)
  call tkspline(knt,nknt,ord-2,nstp,Smat)
  if (ntrt > 1) then
     beta => coef_t(1:(ntrt-1))
  end if
  psi => coef_t(ntrt:ncf)
  tau => coef_t(ncf+1)
  spar => coef_t((ncf+2):(ncf+nstp+3))
  gamma => coef_t(ncf+nstp+5)
! Initial values for parameters
  coef_t=0.0_dp
  tau=0.0_dp
  spar=0.0_dp
  spar(1)=1.0_dp
  gamma=0.0_dp
  d_tpanel=d_tpanel*Panel(npn)
  tdiff=pack(spread(log(d_tpanel),2,nind),.TRUE.)
  where(pack(nijNA,.true.)) tdiff=0.0_dp
  call tkspline(m_tpanel,npan,ord-1,nknt,BB)
  XX(:,ntrt:ncf)=expnd(1:nind).kp.BB
  if (ntrt > 1) then
     XX(:,1:(ntrt-1))=Xmat.kp.expnd(1:npan)
  end if
  XX(pack((/(i,i=1,nXr)/),pack(nijNA,.TRUE.)),:)=0.0_dp
  BBB=XX.tmm.XX
  lognij=pack(nij,.true.)
  where (lognij<0.01_dp)
     lognij=0.01_dp
  elsewhere
     lognij=log(lognij)
  end where
  lognij=lognij-tdiff
end subroutine initialize

subroutine unload
  implicit none
  nullify(beta, psi, spar, tau, gamma)
  deallocate(Bmat, time_grid, coef_t, nij, muij, nijNA, DVD, pDVD, Scr, sqpan, &
       & BBB, lognij, XX, pnind, ipDVD, Smat, mui, ni, dH, Hss, Hss2, fDVD)
  if (ntrt > 1) then
     deallocate(Xmat)
  end if
end subroutine unload

subroutine init_param
  use mytype
  use matrix, only: operator(.mm.), finverse
  implicit none 
  real(dp):: XX2(nXr,ncf), BBB2(ncf,ncf), spr
  integer:: i
  spr=exp(spar(1))
  BBB2=BBB
  do i=(ntrt+ord),ncf
     BBB2(i,i)=BBB2(i,i)+spr
  end do
  BBB2=finverse(BBB2,err)
  if (err /=0) return
  XX2=XX.mm.BBB2
  coef_t(1:ncf)=lognij.mm.XX2
end subroutine init_param

function TauEst(x)
  use mytype
  implicit none
  real(dp), intent(in):: x
  real(dp):: TauEst
  real(dp):: nivar(nind), tmui(nind)
  tau=x
  tmui=1.0_dp+tau*mui
  nivar=mui*tmui
  TauEst=sum((((ni-mui)**2)/nivar + log(nivar)))
end function TauEst

subroutine tau_update
  use mytype
  use fm_util, only: fmin
  implicit none
  real(dp):: fval
  mui=sum(muij,1)
  call fmin(0.0001_dp,tau+1.0_dp,tau,fval,TauEst,1.0e-2_dp) 
end subroutine tau_update

subroutine stepM(n,x,fvec,fjac,ierr)
  use mytype
  implicit none
  integer, intent(in):: n
  real(dp), intent(in):: x(n)
  real(dp), intent(out):: fvec(n)
  real(dp), intent(out):: fjac(n,n)
  integer, intent(out):: ierr
  if (ntrt > 1) then
     beta=x(1:(ntrt-1))
  end if
  psi=x(ntrt:n)
  call update_mu_dmu
  fvec=sum(Scr(1:n,:),2)
  fjac=pDVD
  ierr=err
end subroutine stepM

subroutine stepMT(n,x,fvec,fjac,ierr)
  use mytype
  implicit none
  integer, intent(in):: n
  real(dp), intent(in):: x(n)
  real(dp), intent(out):: fvec(n)
  real(dp), intent(out):: fjac(n,n)
  integer, intent(out):: ierr
  real(dp):: tmui(nind)
  if (ntrt > 1) then
     beta=x(1:(ntrt-1))
  end if
  psi=x(ntrt:n)
  tau=exp(x(n))
  call update_mu_dmu
  tmui=1.0_dp+tau*mui
  fvec(1:(n-1))=sum(Scr(1:(n-1),:),2)
  fvec(n)=sum(tau*((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui) + dH)
  fjac(1:(n-1),1:(n-1))=pDVD
  fjac(n,:)=0.0_dp
  fjac(:,n)=0.0_dp
  fjac(n,n)=sum((((mui/tmui)*tau)**2.0_dp))
  if (tau < 1.0e-7_dp) then
     tau=1.0e-7_dp
     fvec(n)=0.0_dp
     fjac(n,n)=1.0_dp
     return
  end if
  ierr=err
end subroutine stepMT

subroutine mean_update(niter)
  use mytype
  use fm_util, only: newton
  implicit none
  integer, intent(in):: niter
  real(dp):: coeff(ncf), grd(ncf), jaco(ncf,ncf)
  if (ntrt > 1) then
     coeff(1:(ntrt-1))=beta
  end if
  coeff(ntrt:ncf)=psi
  call newton(ncf, coeff, grd, jaco, stepM, niter, err)
end subroutine mean_update

subroutine mean_t_update(niter)
  use mytype
  use fm_util, only: newton
  implicit none
  integer, intent(in):: niter
  real(dp):: coeff(ncf+1), grd(ncf+1), jaco(ncf+1,ncf+1)
  if (ntrt > 1) then
     coeff(1:(ntrt-1))=beta
  end if
  coeff(ntrt:ncf)=psi
  coeff(ncf+1)=log(tau)
  call newton(ncf+1, coeff, grd, jaco, stepMT, niter, err)
end subroutine mean_t_update

subroutine spar_update
  use mytype
  use matrix, only: diag, operator(.mm.), finverse
  implicit none
  real(dp):: tDVD(nBc-ord,nBc-ord)
  ipDVD=finverse(pDVD,err)
  tDVD=ipDVD((ntrt+ord):ncf,(ntrt+ord):ncf)
  spar(1)=sum((psi((ord+1):nBc)*psi((ord+1):nBc)+diag(tDVD)))/real(nBc-ord,dp)
  spar(1)=1.0_dp/(spar(1)+1.0e-7_dp)
  spar(1)=log(spar(1))
end subroutine spar_update

subroutine newton_spar1_step(grd)
  use mytype
  use matrix, only: diag, operator(.mm.), operator(.mmd.), finverse
  implicit none
  real(dp), intent(out):: grd
  real(dp):: tDVD(nBc-ord,nBc-ord), spr, Hss1, tmp(nknt), Hts(nknt,nknt)
  integer:: i, j
  ipDVD=finverse(pDVD,err)
  tDVD=ipDVD((ntrt+ord):ncf,(ntrt+ord):ncf)
  spr=exp(spar(1))
  grd=sum((psi((ord+1):nBc)*psi((ord+1):nBc)+diag(tDVD))*spr)
  tmp=spr
  Hts=tDVD.mmd.tmp
  Hts=Hts.mm.Hts
  Hss1=-sum(diag(Hts))
  Hss1=Hss1+grd
  grd=real(nBc-ord,dp)-grd
  spar(1)=spar(1)+(grd/Hss1)  
  grd=abs(grd)
  if (spar(1) > 5.0_dp) then
     grd=0.0_dp
  end if
end subroutine newton_spar1_step

subroutine newton_spar_step(stsz,scrck)
  use mytype
  use matrix, only: diag, operator(.mm.), operator(.mmd.), finverse, fsymsolve
  implicit none
  real(dp), intent(in):: stsz
  real(dp), intent(out):: scrck
  real(dp):: tDVD(nBc-ord,nBc-ord), spr(nknt), grd(nstp+2), &
       & Hts(nknt,nknt), spr2(nstp+2), tmp(nknt), dSmat(nknt,nstp+2)
  integer:: i, j
  ipDVD=finverse(pDVD,err)
  tDVD=ipDVD((ntrt+ord):ncf,(ntrt+ord):ncf)
  spr=Smat.mm.spar
  spr=exp(spr)
  do i=1,(nstp+2)
     dSmat(:,i)=Smat(:,i)*spr
     grd(i)=sum(Smat(:,i)-((psi((ord+1):nBc)*psi((ord+1):nBc)+diag(tDVD))*dSmat(:,i)))
  end do
  do i=3,(nstp+2)
     grd(i)=grd(i)-2.0_dp*exp(gamma)*spar(i)
  end do
  Hss=0.0_dp
  do i=1,(nstp+2)
     do j=i,(nstp+2)
        tmp=dSmat(:,i)
        Hts=tDVD.mmd.tmp
        Hts=Hts.mm.tDVD
        tmp=dSmat(:,j)
        Hts=Hts.mmd.tmp
        Hss(j,i)=-sum(diag(Hts))
        tmp=dSmat(:,i)*Smat(:,j)
        Hss(j,i)=Hss(j,i)+sum(psi((ord+1):nBc)*psi((ord+1):nBc)*tmp)
        Hts=tDVD.mmd.tmp
        tmp=diag(Hts)
        Hss(j,i)=Hss(j,i)+sum(tmp)
        Hss(i,j)=Hss(j,i)
     end do
  end do
  Hss2=Hss
  do i=3,(nstp+2)
     Hss(i,i)=Hss(i,i)+2.0_dp*exp(gamma)
  end do
  call fsymsolve(Hss,grd,spr2,err)
  if (err /=0) return
  spar=spar+(stsz*spr2)  
  scrck=maxval(abs(grd))
end subroutine newton_spar_step

!!!  Estimate gamma
!subroutine gamma_update
!  use mytype
!  use matrix, only: diag, finverse
!  implicit none
!  real(dp):: dgh(nstp+2)
!  Hss2=finverse(Hss,err)
!  if (err /=0) return
!  dgh=diag(Hss2)
!  gamma=sum(spar(3:(nstp+2))*spar(3:(nstp+2)) + dgh(3:(nstp+2)))/real(nstp,dp)
!  gamma=log(1.0_dp/(gamma+1.0e-7_dp))
!  write(*,*) gamma
!end subroutine gamma_update

function GammaSch(x) result(val)
  use mytype
  use matrix, only: diag, fsymsolve
  implicit none
  real(dp), intent(in):: x
  real(dp):: val
  real(dp):: dgh(nstp+2)
  if (err /= 0) then
     val=1.0e10_dp
     return
  end if
  gamma=x
  call mean_t_spar2_update
  if (err /= 0) then
     val=1.0e10_dp
     return
  end if
  call fsymsolve(Hss,Hss2,Hss2,err)
  if (err /=0) then
     write(*,*) "oops"
     return
  end if
  dgh=diag(Hss2)
  val=sum(dgh)
end function GammaSch

function GammaEst(x) result(val)
  use mytype
  use matrix, only: diag, fsymsolve
  implicit none
  real(dp), intent(in):: x
  real(dp):: val
  real(dp):: dgh(nstp+2)
  if (err /= 0) then
     val=1.0e10_dp
     err=0
     return
  end if
  gamma=x
  call mean_t_spar2_update
  if (err /= 0) then
     val=1.0e10_dp
     return
  end if
  call fsymsolve(Hss,Hss2,Hss2,err)
  if (err /=0) then
     return
  end if
  dgh=diag(Hss2)
  val=abs(sum(dgh)-dof)
end function GammaEst

subroutine gamma_update
  use mytype
  use fm_util, only: fmin
  implicit none
  real(dp):: fval, tgm
  integer:: i
  do i=0,20
     tgm=-real(i*2,dp)
     fval=GammaSch(tgm)
     if(fval > dof) then
        call fmin(tgm,tgm+2.0_dp,gamma,fval,GammaEst,1.0e-2_dp) 
        return
     end if
  end do
end subroutine gamma_update

subroutine mean_spar_update
  use mytype
  implicit none
  real(dp):: scrck
  integer:: i
  do i=1,maxit
     call mean_update(200)
     if (err /= 0) return
     if (i > 10) then
        call newton_spar1_step(scrck)
        if (err /= 0) return    
        if (scrck < 1.0e-5_dp) then
           exit
        end if
     else
        call spar_update
        if (err /= 0) return    
     end if
     if(i==maxit) then
        err=999
     end if
  end do
  call tau_update
end subroutine mean_spar_update

subroutine mean_t_spar_update
  use mytype
  implicit none
  real(dp):: scrck
  integer:: i
  do i=1,maxit
     call mean_t_update(200)
     if (err /= 0) return
     if (i > 10) then
        call newton_spar1_step(scrck)
        if (err /= 0) return    
        if (scrck < 1.0e-5_dp) then
           exit
        end if
     else
        call spar_update
        if (err /= 0) return    
     end if
     if(i==maxit) then
        err=999
     end if
  end do
end subroutine mean_t_spar_update

subroutine mean_t_spar2_update
  use mytype
  implicit none
  real(dp):: scrck
  integer:: i
  do i=1,maxit
     call mean_t_update(200)
     if (err /= 0) return
     if (i <= 10) then
        call newton_spar_step(0.2_dp,scrck)
        if (err /= 0) return 
     else
        call newton_spar_step(1.0_dp,scrck)
        if (err /= 0) return 
     end if
     if (scrck < 1.0e-5_dp) then
        exit
     end if
     if(i==maxit) then
        err=999
     end if
  end do
end subroutine mean_t_spar2_update

subroutine param_cnst_update
  use mytype
  implicit none 
  call mean_spar_update
  if (err /= 0) return
  call mean_t_spar_update
  if (err /= 0) return
end subroutine param_cnst_update

subroutine param_update
  use mytype
  implicit none 
  call mean_spar_update
  if (err /= 0) return
  call mean_t_spar_update
  if (err /= 0) return
  coef_t(ncf+nstp+4)=spar(1)
  call gamma_update
  if (err /= 0) return
  call mean_t_spar2_update
  if (err /= 0) return
end subroutine param_update

subroutine update_mu_dmu
  use mytype
  use fm_util, only: diagadd
  use matrix, only: operator(.mm.), operator(.tmm.), fsymsolve
  implicit none
  integer:: i, j
  real(dp):: tmpint(ntp), Bhold(ntp, nBc), smpsi(nBc-ord), Dmut(npan,ncf), Dmu(npan,nBc), &
       & inmuij(npan,nind), ebeta(nind), spr(nknt), WW(npan,npan), tmui(nind), UU(ncf,nind), &
       & sUU(ncf,nind)
  Dmu=0.0_dp
  if (ntrt > 1) then
     ebeta=exp(Xmat.mm.beta)
  end if
  tmpint=(Bmat.mm.psi)
  tmpint=exp(tmpint)
  do i=1,nBc
     Bhold(:,i)=Bmat(:,i)*tmpint
  end do
  do i=1,npan
     muij(i,:)=sum(tmpint(pnind(1,i):pnind(2,i)))
     Dmu(i,:)=sum(Bhold(pnind(1,i):pnind(2,i),:),1)
  end do
  if (ntrt > 1) then
     do i=1,nind
        muij(:,i)=muij(:,i)*ebeta(i)
     end do
  end if
  where (nijNA)
     muij=0.0_dp
     inmuij=0.0_dp
  elsewhere
     inmuij=1.0_dp/muij
  end where
  DVD=0.0_dp
  mui=sum(muij,1)
  tmui=1.0_dp+tau*mui
  spr=exp(Smat.mm.spar)
  smpsi=(spr*psi((ord+1):nBc))/real(nind,dp)
  do i=1,nind
     if (ntrt > 1) then
        do j=1,(ntrt-1)
           Dmut(:,j)=Xmat(i,j)*muij(:,i)
        end do
        do j=ntrt,ncf
           Dmut(:,j)=Dmu(:,j-ntrt+1)*ebeta(i)
        end do
     else
        Dmut=Dmu
     end if
     WW=-tau/tmui(i)
     call diagadd(WW,inmuij(:,i))
     Dmut(pack(sqpan,nijNA(:,i)),:)=0.0_dp
     Scr(1:ncf,i)=((WW.mm.(nij(:,i)-muij(:,i))).mm.Dmut)
     Scr((ntrt+ord):ncf,i)=Scr((ntrt+ord):ncf,i)-smpsi
     fDVD(i,:,:)=((Dmut.tmm.WW).mm.Dmut)
     DVD=DVD+fDVD(i,:,:)
     do j=(ntrt+ord),ncf
        fDVD(i,j,j)=fDVD(i,j,j)+(spr(j-ntrt-ord+1)/real(nind,dp))
     end do
     UU(:,i)=sum(Dmut,1)/tmui(i)
  end do
  pDVD=sum(fDVD,1)
  call fsymsolve(pDVD,UU,sUU,err)
  if (err /=0) return
  dH=sum(UU*sUU,1)
end subroutine update_mu_dmu

! Robust Variance Estimate (White)
!function SandVar() result(VarCov)
!  use mytype
!  use matrix, only: operator(.mmt.), fsymsolve, diag
!  implicit none
!  real(dp):: VarCov(ncf+1,ncf+1), Scb(ncf+1,nind), tmui(nind), tDVD(ncf,ncf)
!  VarCov=0.0_dp
!  call update_mu_dmu
!  call fsymsolve(pDVD,DVD,tDVD,err)
!  if (err /= 0) return
!  eff_df=sum(diag(tDVD))
!  Scb(1:ncf,:)=Scr
!  VarCov(1:ncf,1:ncf)=pDVD
!  tmui=1.0_dp+tau*mui
!  Scb(ncf+1,:)=(((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui)+dH)
!  VarCov(ncf+1,ncf+1)=sum(((mui/tmui)**2.0_dp))
!  call fsymsolve(VarCov,Scr,Scb,err)
!  if (err/=0) return
!  VarCov=Scb.mmt.Scb
!end function SandVar

function JackVar() result(VarCov)
  use mytype
  use matrix, only: operator(.mmt.), fsymsolve, diag
  implicit none
  real(dp):: VarCov(ncf+1,ncf+1), Scb(ncf+1,nind), tmui(nind), tsc(nind), tjc(nind), &
       & grd(ncf+1), fjc(ncf+1,ncf+1), tDVD(ncf,ncf)
  integer:: i, indx(nind)
  VarCov=0.0_dp
  call update_mu_dmu
  call fsymsolve(pDVD,DVD,tDVD,err)
  if (err /= 0) return
  indx=[(i,i=1,nind)]
  eff_df=sum(diag(tDVD))
  tmui=1.0_dp+tau*mui
  tsc=(((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui) + dH)
  tjc=((mui/tmui)**2.0_dp)
  do i=1,nind
     fjc=0.0_dp
     grd(1:ncf)=sum(Scr(:,indx(1:(nind-1))),2)
     grd(ncf+1)=sum(tsc(indx(1:(nind-1))))
     fjc(1:ncf,1:ncf)=sum(fDVD(indx(1:(nind-1)),:,:),1)
     fjc(ncf+1,ncf+1)=sum(tjc(indx(1:nind-1)))
     call fsymsolve(fjc,grd,Scb(:,i),err)
     if (err/=0) return
     indx=[indx(2:nind),indx(1)]
  end do
  VarCov=((nXr-eff_df-1.0_dp)/nXr)*(Scb.mmt.Scb)
end function JackVar

subroutine bootstrp(tw,bcoefs,nbt)
  use mytype
  use matrix, only: fsymsolve
  implicit none
  integer, intent(in):: tw(nbt,nind), nbt
  real(dp), intent(out):: bcoefs(ncf+1,nbt)
  real(dp):: Scb(ncf,nind), grd(ncf+1), fjc(ncf+1,ncf+1), tmui(nind), onij(npan,nind), &
       & oni(nind)
  real(dp), allocatable:: Xm(:,:)
  logical:: doCov
  integer:: i
  doCov=.FALSE.
  if (ntrt > 1) then
     allocate(Xm(nind,ntrt-1))
     doCov=.TRUE.
  end if
  onij=nij
  oni=ni
  if (doCov) Xm=Xmat
  do i=1,nbt
     fjc=0.0_dp
     nij=onij(:,tw(i,:))
     ni=oni(tw(i,:))
     if (doCov) Xmat=Xm(tw(i,:),:)
     call update_mu_dmu
     tmui=1.0_dp+tau*mui
     grd(1:ncf)=sum(Scr,2)
     grd(ncf+1)=sum((((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui) + dH))
     fjc(1:ncf,1:ncf)=sum(fDVD(tw(i,:),:,:),1)
     fjc(ncf+1,ncf+1)=sum((mui/tmui)**2.0_dp)
     call fsymsolve(fjc,grd,grd,err)
     if (err/=0) return
     bcoefs(:,i)=coef_t(1:(ncf+1))+grd
  end do
  nij=onij
  ni=oni
  if (doCov) then
     Xmat=Xm
     deallocate(Xm)
  end if
end subroutine bootstrp

function calc_pql(sgsp) result(val)
  use mytype
  use fm_util, only: diagadd
  use matrix, only: operator(.mm.), ldet
  implicit none 
  logical, intent(in):: sgsp
  real(dp):: val, tmui(nind), res(npan), inmuij(npan,nind), spr(nknt), lgd, &
       & WW(npan,npan)
  integer:: i
  val=0.0_dp
  where (nijNA)
     muij=0.0_dp
     inmuij=0.0_dp
  elsewhere
     inmuij=1.0_dp/muij
  end where
  mui=sum(muij,1)
  tmui=1.0_dp+tau*mui
  do i=1,nind
     WW=-tau/tmui(i)
     call diagadd(WW,inmuij(:,i))
     res=nij(:,i)-muij(:,i)
     val=val+sum(res*(WW.mm.res))     
  end do
  spr=exp(Smat.mm.spar)
  val=val+sum(spr*psi((ord+1):nBc)*psi((ord+1):nBc))-sum(log(spr))-real(nstp,dp)*gamma
  do i=1,ncf
     pDVD(i,i)=pDVD(i,i)+1.0e-7_dp
  end do
  call ldet(pDVD,ncf,lgd)
  val=val+lgd
  if (sgsp) return
  do i=1,(nstp+2)
     Hss(i,i)=Hss(i,i)+1.0e-7_dp
  end do
  call ldet(Hss,nstp+2,lgd)
  val=val+lgd
end function calc_pql

subroutine fmNHPP(panel,datij,Xm,gmm,tw,bcoefs,coefs,VrCov,pql,edf,npn,nsub,ntr,nkn,nst,nbt,sgsp,info)
  use mytype
  implicit none
  integer, intent(in):: npn, nsub, ntr, nkn, nst, nbt
  logical, intent(in):: sgsp
  real(dp), intent(in):: panel(npn), datij(npn,nsub), gmm
  real(dp), intent(in):: Xm(nsub,ntr)
  integer, intent(in):: tw(nbt,nsub)
  real(dp), intent(out):: bcoefs(ntr+nkn+3,nbt), coefs(ntr+nkn+nst+7), edf, pql
  real(dp), intent(out):: VrCov((ntr+nkn+3),(ntr+nkn+3))
  integer, intent(out):: info
  call initialize(datij,Xm,panel,gmm,npn,nsub,ntr,nkn,nst)
  call init_param
  if (err /= 0) then
     call unload
     info=err
     return
  end if
  if (err == 0) then
     if (sgsp) then
        call param_cnst_update
        if (err /=0) then
           call unload
           info=err        
           return
        end if
     else
        call param_update
        if (err /=0) then
           call unload
           info=err        
           return
        end if
     end if
     VrCov=JackVar()
     pql=calc_pql(sgsp)
     edf=eff_df+1
     coefs=coef_t
     call bootstrp(tw,bcoefs,nbt)
  else
     info=err
  end if
  call unload
end subroutine fmNHPP

end module fm_NHPP_mod

subroutine R_fmNHPP(panel,datij,Xm,gmm,tw,bcoefs,coefs,VrCov,pql,edf,npn,nsub,ntr,nkn,nst,nbt,sgsp,info)
  use mytype
  use fm_NHPP_mod
  implicit none
  integer, intent(in):: npn, nsub, ntr, nkn, nst, nbt
  logical, intent(in):: sgsp
  real(dp), intent(in):: panel(npn), datij(npn,nsub), gmm
  real(dp), intent(in):: Xm(nsub,ntr)
  integer, intent(in):: tw(nbt,nsub)
  real(dp), intent(out):: bcoefs(ntr+nkn+3,nbt), coefs(ntr+nkn+nst+7), edf, pql
  real(dp), intent(out):: VrCov((ntr+nkn+3),(ntr+nkn+3))
  integer, intent(out):: info
  call fmNHPP(panel,datij,Xm,gmm,tw,bcoefs,coefs,VrCov,pql,edf,npn,nsub,ntr,nkn,nst,nbt,sgsp,info)
end subroutine R_fmNHPP
